
# This file was generated by 'versioneer.py' (0.28) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-07T14:13:01-0600",
 "dirty": false,
 "error": null,
 "full-revisionid": "6dc5479eca87556eb9eb1100bc0cd0f6bc9d3a13",
 "version": "0.14.2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
